/*
 * Copyright 2021 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.android.fhir.datacapture.contrib.views.barcode.mlkit.md

import android.content.ContentResolver
import android.content.Context
import android.content.pm.PackageManager
import android.content.res.Configuration
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.Canvas
import android.graphics.Matrix
import android.graphics.Paint
import android.graphics.PorterDuff
import android.graphics.PorterDuffXfermode
import android.graphics.Rect
import android.graphics.RectF
import android.graphics.YuvImage
import android.hardware.Camera
import android.net.Uri
import androidx.exifinterface.media.ExifInterface
import com.google.android.fhir.datacapture.contrib.views.barcode.mlkit.md.camera.CameraSizePair
import com.google.mlkit.vision.barcode.BarcodeScanning
import com.google.mlkit.vision.common.InputImage
import java.io.ByteArrayOutputStream
import java.io.IOException
import java.io.InputStream
import java.nio.ByteBuffer
import java.util.ArrayList
import kotlin.math.abs
import timber.log.Timber

/** Utility class to provide helper methods. */
object Utils {

  /**
   * If the absolute difference between aspect ratios is less than this tolerance, they are
   * considered to be the same aspect ratio.
   */
  const val ASPECT_RATIO_TOLERANCE = 0.01f

  internal const val REQUEST_CODE_PHOTO_LIBRARY = 1

  private fun getRequiredPermissions(context: Context): Array<String> {
    return try {
      val info =
        context.packageManager.getPackageInfo(context.packageName, PackageManager.GET_PERMISSIONS)
      val ps = info.requestedPermissions
      if (ps != null && ps.isNotEmpty()) ps else arrayOf()
    } catch (e: Exception) {
      arrayOf()
    }
  }

  fun isPortraitMode(context: Context): Boolean =
    context.resources.configuration.orientation == Configuration.ORIENTATION_PORTRAIT

  /**
   * Generates a list of acceptable preview sizes. Preview sizes are not acceptable if there is not
   * a corresponding picture size of the same aspect ratio. If there is a corresponding picture size
   * of the same aspect ratio, the picture size is paired up with the preview size.
   *
   * This is necessary because even if we don't use still pictures, the still picture size must be
   * set to a size that is the same aspect ratio as the preview size we choose. Otherwise, the
   * preview images may be distorted on some devices.
   */
  fun generateValidPreviewSizeList(camera: Camera): List<CameraSizePair> {
    val parameters = camera.parameters
    val supportedPreviewSizes = parameters.supportedPreviewSizes
    val supportedPictureSizes = parameters.supportedPictureSizes
    val validPreviewSizes = ArrayList<CameraSizePair>()
    for (previewSize in supportedPreviewSizes) {
      val previewAspectRatio = previewSize.width.toFloat() / previewSize.height.toFloat()

      // By looping through the picture sizes in order, we favor the higher resolutions.
      // We choose the highest resolution in order to support taking the full resolution
      // picture later.
      for (pictureSize in supportedPictureSizes) {
        val pictureAspectRatio = pictureSize.width.toFloat() / pictureSize.height.toFloat()
        if (abs(previewAspectRatio - pictureAspectRatio) < ASPECT_RATIO_TOLERANCE) {
          validPreviewSizes.add(CameraSizePair(previewSize, pictureSize))
          break
        }
      }
    }

    // If there are no picture sizes with the same aspect ratio as any preview sizes, allow all of
    // the preview sizes and hope that the camera can handle it.  Probably unlikely, but we still
    // account for it.
    if (validPreviewSizes.isEmpty()) {
      Timber.w("No preview sizes have a corresponding same-aspect-ratio picture size.")
      for (previewSize in supportedPreviewSizes) {
        // The null picture size will let us know that we shouldn't set a picture size.
        validPreviewSizes.add(CameraSizePair(previewSize, null))
      }
    }

    return validPreviewSizes
  }

  fun getCornerRoundedBitmap(srcBitmap: Bitmap, cornerRadius: Int): Bitmap {
    val dstBitmap = Bitmap.createBitmap(srcBitmap.width, srcBitmap.height, Bitmap.Config.ARGB_8888)
    val canvas = Canvas(dstBitmap)
    val paint = Paint()
    paint.isAntiAlias = true
    val rectF = RectF(0f, 0f, srcBitmap.width.toFloat(), srcBitmap.height.toFloat())
    canvas.drawRoundRect(rectF, cornerRadius.toFloat(), cornerRadius.toFloat(), paint)
    paint.xfermode = PorterDuffXfermode(PorterDuff.Mode.SRC_IN)
    canvas.drawBitmap(srcBitmap, 0f, 0f, paint)
    return dstBitmap
  }

  /** Convert NV21 format byte buffer to bitmap. */
  fun convertToBitmap(data: ByteBuffer, width: Int, height: Int, rotationDegrees: Int): Bitmap? {
    data.rewind()
    val imageInBuffer = ByteArray(data.limit())
    data.get(imageInBuffer, 0, imageInBuffer.size)
    try {
      val image = YuvImage(imageInBuffer, InputImage.IMAGE_FORMAT_NV21, width, height, null)
      val stream = ByteArrayOutputStream()
      image.compressToJpeg(Rect(0, 0, width, height), 80, stream)
      val bmp = BitmapFactory.decodeByteArray(stream.toByteArray(), 0, stream.size())
      stream.close()

      // Rotate the image back to straight.
      val matrix = Matrix()
      matrix.postRotate(rotationDegrees.toFloat())
      return Bitmap.createBitmap(bmp, 0, 0, bmp.width, bmp.height, matrix, true)
    } catch (e: java.lang.Exception) {
      Timber.e("Error: " + e.message)
    }
    return null
  }

  @Throws(IOException::class)
  internal fun loadImage(context: Context, imageUri: Uri, maxImageDimension: Int): Bitmap? {
    var inputStreamForSize: InputStream? = null
    var inputStreamForImage: InputStream? = null
    try {
      inputStreamForSize = context.contentResolver.openInputStream(imageUri)
      var opts = BitmapFactory.Options()
      opts.inJustDecodeBounds = true
      BitmapFactory.decodeStream(inputStreamForSize, null, opts)
      /* outPadding= */ val inSampleSize =
        Math.max(opts.outWidth / maxImageDimension, opts.outHeight / maxImageDimension)

      opts = BitmapFactory.Options()
      opts.inSampleSize = inSampleSize
      inputStreamForImage = context.contentResolver.openInputStream(imageUri)
      val decodedBitmap = BitmapFactory.decodeStream(inputStreamForImage, null, opts)
      /* outPadding= */ return maybeTransformBitmap(
        context.contentResolver,
        imageUri,
        decodedBitmap
      )
    } finally {
      inputStreamForSize?.close()
      inputStreamForImage?.close()
    }
  }

  private fun maybeTransformBitmap(resolver: ContentResolver, uri: Uri, bitmap: Bitmap?): Bitmap? {
    val matrix: Matrix? =
      when (getExifOrientationTag(resolver, uri)) {
        ExifInterface.ORIENTATION_UNDEFINED, ExifInterface.ORIENTATION_NORMAL ->
          // Set the matrix to be null to skip the image transform.
          null
        ExifInterface.ORIENTATION_FLIP_HORIZONTAL -> Matrix().apply { postScale(-1.0f, 1.0f) }
        ExifInterface.ORIENTATION_ROTATE_90 -> Matrix().apply { postRotate(90f) }
        ExifInterface.ORIENTATION_TRANSPOSE -> Matrix().apply { postScale(-1.0f, 1.0f) }
        ExifInterface.ORIENTATION_ROTATE_180 -> Matrix().apply { postRotate(180.0f) }
        ExifInterface.ORIENTATION_FLIP_VERTICAL -> Matrix().apply { postScale(1.0f, -1.0f) }
        ExifInterface.ORIENTATION_ROTATE_270 -> Matrix().apply { postRotate(-90.0f) }
        ExifInterface.ORIENTATION_TRANSVERSE ->
          Matrix().apply {
            postRotate(-90.0f)
            postScale(-1.0f, 1.0f)
          }
        else ->
          // Set the matrix to be null to skip the image transform.
          null
      }

    return if (matrix != null) {
      Bitmap.createBitmap(bitmap!!, 0, 0, bitmap.width, bitmap.height, matrix, true)
    } else {
      bitmap
    }
  }

  private fun getExifOrientationTag(resolver: ContentResolver, imageUri: Uri): Int {
    if (ContentResolver.SCHEME_CONTENT != imageUri.scheme &&
        ContentResolver.SCHEME_FILE != imageUri.scheme
    ) {
      return 0
    }

    var exif: ExifInterface? = null
    try {
      resolver.openInputStream(imageUri)?.use { inputStream -> exif = ExifInterface(inputStream) }
    } catch (e: IOException) {
      Timber.e("Failed to open file to read rotation meta data: $imageUri", e)
    }

    return if (exif != null) {
      exif!!.getAttributeInt(ExifInterface.TAG_ORIENTATION, ExifInterface.ORIENTATION_NORMAL)
    } else {
      ExifInterface.ORIENTATION_UNDEFINED
    }
  }

  fun getBarcodeScanningClient() = BarcodeScanning.getClient()
}
